import json
import logging
from os import environ, path
from typing import Any, Dict, List, Optional, Union

import yaml

from .node import Node
from .vault import Vault

logger = logging.getLogger(__name__)


class Config:
    def __init__(self, files: Optional[List[str]] = None,
                 passkey: Optional[str] = None):
        self._files: List[str] = []
        self._configuration: Dict[str, Any] = {}
        self._passkey: Optional[bytes] = None

        if passkey:
            self._passkey = passkey.encode("utf-8")
        elif environ.get("VAULT_PASSKEY"):
            self._passkey = environ["VAULT_PASSKEY"].encode("utf-8")
        if files:
            for config_file in files:
                self.load(config_file)

    def reload(self):
        for config_file in self._files:
            self.load(config_file)

    def load(self, config_file: str, silent=False):
        logger.info("Trying to load configuration from: %s", config_file)
        if not path.exists(config_file):
            (logger.warn if silent else logger.critical)(
                "Missing config file: %s", config_file)
            if not silent:
                raise FileNotFoundError()
            return False
        if config_file.endswith((".yml", ".yaml")):
            self._parse_yml(config_file)
        elif config_file.endswith(".json"):
            self._parse_json(config_file)
        else:
            return False
        if config_file not in self._files:
            self._files.append(config_file)
        return True

    def find(self, path: str, default=None):
        cursor = self._configuration
        for key in path.split("."):
            if isinstance(cursor, dict) and key in cursor:
                cursor = cursor[key]
            else:
                return default
        return cursor

    def _parse_yml(self, config_file):
        with open(config_file) as fd:
            payload = fd.read()
        data = yaml.safe_load(self._strip_encrytion(payload))
        self._deep_update(data)

    def _parse_json(self, config_file):
        with open(config_file) as fd:
            payload = fd.read()
        data = json.loads(self._strip_encrytion(payload))
        self._deep_update(data)

    def _strip_encrytion(self, payload: str) -> str:
        if payload.startswith(Vault.prefix()):
            if self._passkey is None:
                raise RuntimeError("No decryption key provided")
            vault = Vault(self._passkey)
            decrypted = vault.decrypt(payload)
            return decrypted
        else:
            return payload

    def _deep_update(self, source: dict, target: Optional[dict] = None):
        if target is None:
            target = self._configuration
        if not source:
            return source
        for key, value in source.items():
            if isinstance(value, dict):
                if key not in target or not isinstance(target[key], dict):
                    target[key] = {}
                self._deep_update(value, target[key])
            else:
                target[key] = value

    def __getattribute__(self, attr: str):
        methods = ("reload", "load", "find")
        if attr.startswith("_") or attr in methods:
            return super().__getattribute__(attr)

        value: Union[Dict, Node, None] = self._configuration
        keys = attr.split(".")
        for key in keys:
            if value is not None and key in value:
                value = value[key]
                if not isinstance(value, dict):
                    return value
                value = Node(value)
            else:
                value = None
        return value



from base64 import b64decode, b64encode

from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes


class Vault:
    magic = "$Vault"
    separator = ";"

    def __init__(self, passkey: bytes):
        self.passkey = passkey

    def encrypt(self, payload: str) -> str:
        nonce = get_random_bytes(12)
        cipher = AES.new(self.passkey, AES.MODE_GCM, nonce=nonce)
        data, tag = cipher.encrypt_and_digest(payload.encode("utf-8"))
        return self.separator.join([
            self.magic,
            b64encode(nonce).decode("utf-8"),
            b64encode(data).decode("utf-8"),
            b64encode(tag).decode("utf-8")])

    def decrypt(self, payload: str) -> str:
        magic, nonce, data, tag, *_meta = payload.split(self.separator)
        if magic == self.magic and nonce and data:
            cipher = AES.new(self.passkey, AES.MODE_GCM,
                             nonce=b64decode(nonce))
            return cipher.decrypt_and_verify(
                b64decode(data), b64decode(tag)).decode("utf-8")
        raise RuntimeError("Couldn't decrypt payload")

    @staticmethod
    def prefix():
        return f"{Vault.magic}{Vault.separator}"



#!/usr/bin/python3
import logging
import tempfile
from argparse import ArgumentParser
from os import environ, path, unlink
from subprocess import Popen
from sys import stderr

from confit2canard import Vault

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler(stderr)
logger.addHandler(handler)


def main(configuration, passkey: str = ""):
    if not path.exists(path.abspath(path.dirname(configuration))):
        logger.error("Parent folder for %s doesn't exist", configuration)
        return
    editor = environ.get("EDITOR")
    vault = Vault((environ.get("VAULT_PASSKEY") or passkey).encode("utf-8"))
    if not editor:
        logging.error("EDITOR environment variable not set")
        return

    extension = "." + configuration.split(".")[-1]
    old = ""
    with tempfile.NamedTemporaryFile(delete=False, suffix=extension,
                                     mode="w") as tfile:
        if path.exists(configuration):
            logger.info("Reading configuration from %s", configuration)
            with open(configuration, "r") as fd:
                buffer = fd.read()
                if buffer:
                    try:
                        old = vault.decrypt(buffer)
                    except ValueError:
                        logger.error("Couldn't decrypt, the key might not "
                                     "be valid.")
                        tfile.close()
                        unlink(tfile.name)
                        return
                    tfile.write(old)
        tfile.close()
        process = Popen([editor, tfile.name])
        process.wait()

        if process.returncode != 0:
            logging.error("An error occurred with editor %s", editor)
            return

        with open(tfile.name, "r") as fd:
            payload = fd.read()

    if len(payload) == 0:
        logger.info("Nothing to write")
        return
    if payload != "":
        payload = vault.encrypt(payload)
    try:
        with open(configuration, "w+") as fd:
            fd.write(payload)
            logger.info("Writing %d bytes", len(payload))
    except PermissionError:
        logger.error("Couldn't write configuration to file: %s", payload)


if __name__ == "__main__":
    args = ArgumentParser("Confit2Canard")
    args.add_argument("config", action="store",
                      help="The path to the file to edit")

    parsed = args.parse_args()
    main(parsed.config)



class Node:
    def __init__(self, data: dict):
        self._keys = []
        for key, value in data.items():
            setattr(self, key, value)
            self._keys.append(key)

    def __getattribute__(self, attr: str):
        value = super().__getattribute__(attr)
        if isinstance(value, dict):
            value = Node(value)
        return value

    def __getitem__(self, item):
        return getattr(self, item, None)

    def __iter__(self):
        for key in self._keys:
            yield key
